#include "node.hpp"

#include <vector>
#include <cmath>

/* 
Node::Node() {
    root_ = nullptr;
    parent_ = nullptr;
    particles_ = nullptr;
    nParticles_ = 0;
    depth_ = 0;
    size_ = 0;
    for (unsigned i=0; i<3; ++i) center_[i] = 0;
    localParticles_ = nullptr;
    nLocalParticles_ = 0;
    for (unsigned i=0; i<8; ++i) children_[i] = nullptr;
}
 */

Node::Node(Node* root, Node* parent, double* masses, double* particles, unsigned nParticles, unsigned depth, 
        double size, double center[3], List& localParticles, unsigned nLocalParticles) {
    parent_ = parent;
    particles_ = particles;
    nParticles_ = nParticles;

    depth_ = depth;
    size_ = size;

    for (unsigned i=0; i<3; ++i) center_[i] = center[i];

    localParticles_ = localParticles;
    nLocalParticles_ = nLocalParticles;

    mass_ = 0;
    for (unsigned i=0; i<nLocalParticles; ++i) {
        mass_ += masses_[localParticles_[i]];
    }

    // TODO
    // allocate children
    // check if child contains no particles, in that case set child to nullptr
    // check if node is leaf (it contains a single particle), in that case all children are nullptr
    if (nLocalParticles <= 1) {
        isLeaf_ = true;
        for (unsigned i=0; i<8; ++i)
            children_[i] = nullptr;
    } else {
        isLeaf_ = false;
        
        // for each local particle, determine which octant it is in
        List* particleDivision = new List[nLocalParticles];
        for (unsigned i=0; i<nLocalParticles; ++i) {
            particleDivision[getOctant(localParticles_[i])].push_back(i);
        }

        // for each octant, create a child node
        for (unsigned i=0; i<8; ++i) {
            if (particleDivision[i].size() == 0) {
                children_[i] = nullptr;
            } else {
                double childSize = size_/2;
                double childCenter[3];
                for (unsigned j=0; j<3; ++j) {
                    childCenter[j] = center_[j] + (i & (1 << j) ? childSize : -childSize);
                }
                children_[i] = new Node(root, this, masses, particles, nParticles, depth+1, 
                        childSize, childCenter, particleDivision[i], particleDivision[i].size());
            }
        }
        delete[] particleDivision;
    }
}

Node::~Node() {
    for (unsigned i=0; i<8; ++i) {
        delete children_[i];
    }
}

Node& Node::operator=(const Node& other) {
    root_ = other.root_;
    parent_ = other.parent_;
    particles_ = other.particles_;
    nParticles_ = other.nParticles_;
    depth_ = other.depth_;
    size_ = other.size_;
    for (unsigned i=0; i<3; ++i) center_[i] = other.center_[i];
    localParticles_ = other.localParticles_;
    nLocalParticles_ = other.nLocalParticles_;
    mass_ = other.mass_;
    for (unsigned i=0; i<8; ++i) children_[i] = other.children_[i];
    return *this;
}

void Node::update(List& localParticles) {
    // TODO
    // update center of mass
    // maybe subdivide? 
    // update this node
    // update children

    // for each local particle, determine which octant it is in
    List* particleDivision = new List[nLocalParticles_];
    for (unsigned i=0; i<nLocalParticles_; ++i) {
        particleDivision[getOctant(localParticles_[i])].push_back(i);
    }

    isLeaf_ = true;
    for (unsigned i=0; i<8; ++i) {
        if (children_[i] != nullptr) {
            // if child is not a nullptr and particleDivision[i] is empty, delete the child
            // otherwise update the child
            if (particleDivision[i].size() == 0) {
                delete children_[i];
                children_[i] = nullptr;
            } else {
                children_[i]->update(particleDivision[i]);
                isLeaf_ = false;
            }
        } else {
            // if child is a nullptr and particleDivision[i] is not empty, create a new child
            if (particleDivision[i].size() != 0) {
                double childSize = size_/2;
                double childCenter[3];
                for (unsigned j=0; j<3; ++j) {
                    childCenter[j] = center_[j] + (i & (1 << j) ? childSize : -childSize);
                }
                children_[i] = new Node(root_, this, masses_, particles_, nParticles_, depth_+1, 
                        childSize, childCenter, particleDivision[i], particleDivision[i].size());
                isLeaf_ = false;
            }
        }
    }

    delete[] particleDivision;
}

double* Node::calculateForce(unsigned particle) {
    // calculate theta
    double dx = center_[0] - particles_[3*particle + 0];
    double dy = center_[1] - particles_[3*particle + 1];
    double dz = center_[2] - particles_[3*particle + 2];
    double lambda = std::sqrt(dx*dx + dy*dy + dz*dz);
    double theta = size_/lambda;

    // TODO
    // If angle too large, call this function for every child which is not a nullptr. 
    //      If the node is a leaf, calculate force on particle from particles in this node. 
    //      If the node is not a leaf, call this function for every child which is not a nullptr.
    // If angle small enough, calculate force on particle from this node.
    if (theta > theta0_) {
        if (isLeaf_) {
            // TODO
        } else {
            // TODO
        }
    } else {
        // TODO
    }

    return nullptr;
}

void Node::calculateCenterOfMass() {
    for (unsigned j=0; j<3; ++j)
        centerOfMass_[j] = 0;
    
    for (unsigned i : localParticles_)
        for (unsigned j=0; j<3; ++j)
            centerOfMass_[j] += particles_[3*i + j] / nLocalParticles_;
}

unsigned Node::getOctant(unsigned particle) {
    double px = particles_[localParticles_[0]];
    double py = particles_[localParticles_[1]];
    double pz = particles_[localParticles_[2]];
    if (px < center_[0]) {
        if (py < center_[1]) {
            if (pz < center_[2]) {
                return 0;
            } else {
                return 1;
            }
        } else {
            if (pz < center_[2]) {
                return 2;
            } else {
                return 3;
            }
        }
    } else {
        if (py < center_[1]) {
            if (pz < center_[2]) {
                return 4;
            } else {
                return 5;
            }
        } else {
            if (pz < center_[2]) {
                return 6;
            } else {
                return 7;
            }
        }
    }
}